@@ -20,132 +20,22 @@ along with jpegxl-rs.  If not, see <https://www.gnu.org/licenses/>.
 use std::{marker::PhantomData, mem::MaybeUninit, ops::Deref, ptr::null};
 
 #[allow(clippy::wildcard_imports)]
-use jpegxl_sys::{
-    color_encoding::JxlColorEncoding,
-    encode::*,
-    types::{JxlEndianness, JxlPixelFormat},
-};
+use jpegxl_sys::encode::*;
 
 use crate::{
     common::PixelType, errors::EncodeError, memory::MemoryManager, parallel::JxlParallelRunner,
 };
 
-/// Encoding speed
-#[derive(Debug, Clone, Copy)]
-pub enum EncoderSpeed {
-    /// Fastest, 1
-    Lightning = 1,
-    /// 2
-    Thunder = 2,
-    /// 3
-    Falcon = 3,
-    /// 4
-    Cheetah,
-    /// 5
-    Hare,
-    /// 6
-    Wombat,
-    /// 7, default
-    Squirrel,
-    /// 8
-    Kitten,
-    /// Slowest, 9
-    Tortoise,
-}
+mod options;
+pub use options::*;
 
-impl std::default::Default for EncoderSpeed {
-    fn default() -> Self {
-        Self::Squirrel
-    }
-}
+mod metadata;
+pub use metadata::*;
 
-/// Encoding color profile
-#[derive(Debug, Clone, Copy)]
-pub enum ColorEncoding {
-    /// SRGB, default for uint pixel types
-    Srgb,
-    /// Linear SRGB, default for float pixel types
-    LinearSrgb,
-    /// SRGB, images with only luma channel
-    SrgbLuma,
-    /// Linear SRGB with only luma channel
-    LinearSrgbLuma,
-}
+mod frame;
+pub use frame::*;
 
-impl From<ColorEncoding> for JxlColorEncoding {
-    fn from(val: ColorEncoding) -> Self {
-        use ColorEncoding::{LinearSrgb, LinearSrgbLuma, Srgb, SrgbLuma};
-
-        let mut color_encoding = MaybeUninit::uninit();
-
-        unsafe {
-            match val {
-                Srgb => JxlColorEncodingSetToSRGB(color_encoding.as_mut_ptr(), false),
-                LinearSrgb => JxlColorEncodingSetToLinearSRGB(color_encoding.as_mut_ptr(), false),
-                SrgbLuma => JxlColorEncodingSetToSRGB(color_encoding.as_mut_ptr(), true),
-                LinearSrgbLuma => {
-                    JxlColorEncodingSetToLinearSRGB(color_encoding.as_mut_ptr(), true);
-                }
-            }
-            color_encoding.assume_init()
-        }
-    }
-}
-
-/// A frame for the encoder, consisting of the pixels and its options
-pub struct EncoderFrame<'data, T: PixelType> {
-    data: &'data [T],
-    num_channels: Option<u32>,
-    endianness: Option<JxlEndianness>,
-    align: Option<usize>,
-}
-
-impl<'data, T: PixelType> EncoderFrame<'data, T> {
-    /// Create a default frame from the data.
-    ///
-    /// Use RGB(3) channels, native endianness and no alignment.
-    pub fn new(data: &'data [T]) -> Self {
-        Self {
-            data,
-            num_channels: None,
-            endianness: None,
-            align: None,
-        }
-    }
-
-    /// Set the number of channels of the source.
-    ///
-    /// _Note_: If you want to use alpha channel, add here
-    #[must_use]
-    pub fn num_channels(mut self, value: u32) -> Self {
-        self.num_channels = Some(value);
-        self
-    }
-
-    /// Set the endianness of the source.
-    #[must_use]
-    pub fn endianness(mut self, value: JxlEndianness) -> Self {
-        self.endianness = Some(value);
-        self
-    }
-
-    /// Set the align of the source.
-    /// Align scanlines to a multiple of align bytes, or 0 to require no alignment at all
-    #[must_use]
-    pub fn align(mut self, value: usize) -> Self {
-        self.align = Some(value);
-        self
-    }
-
-    fn pixel_format(&self) -> JxlPixelFormat {
-        JxlPixelFormat {
-            num_channels: self.num_channels.unwrap_or(3),
-            data_type: T::pixel_type(),
-            endianness: self.endianness.unwrap_or(JxlEndianness::Native),
-            align: self.align.unwrap_or(0),
-        }
-    }
-}
+// MARK: Utility types
 
 /// Encoder result
 pub struct EncoderResult<U: PixelType> {
@@ -162,6 +52,8 @@ impl<U: PixelType> Deref for EncoderResult<U> {
     }
 }
 
+// MARK: Encoder
+
 /// JPEG XL Encoder
 #[derive(Builder)]
 #[builder(build_fn(skip, error = "None"))]
@@ -231,6 +123,9 @@ pub struct JxlEncoder<'prl, 'mm> {
     /// Default: `None`, indicating single thread execution
     pub parallel_runner: Option<&'prl dyn JxlParallelRunner>,
 
+    /// Whether box is used in encoder
+    use_box: bool,
+
     /// Set memory manager
     #[allow(dead_code)]
     memory_manager: Option<&'mm dyn MemoryManager>,
@@ -273,12 +168,20 @@ impl<'prl, 'mm> JxlEncoderBuilder<'prl, 'mm> {
             init_buffer_size,
             color_encoding: self.color_encoding.unwrap_or(ColorEncoding::Srgb),
             parallel_runner: self.parallel_runner.flatten(),
+            use_box: self.use_box.unwrap_or_default(),
             memory_manager: mm,
         })
     }
+
+    /// Set the `quality` parameter from a JPEG-style quality factor (0-100, higher is better
+    /// quality).
+    pub fn set_jpeg_quality(&mut self, quality: f32) {
+        // SAFETY: the C API has no safety requirements.
+        self.quality(unsafe { JxlEncoderDistanceFromQuality(quality) });
+    }
 }
 
-// Private helper functions
+// MARK: Private helper functions
 impl JxlEncoder<'_, '_> {
     /// Error mapping from underlying C const to [`EncodeError`] enum
     #[cfg_attr(coverage_nightly, coverage(off))]
@@ -453,7 +356,7 @@ impl JxlEncoder<'_, '_> {
     }
 }
 
-// Public interface
+// MARK: Public interface
 impl<'prl, 'mm> JxlEncoder<'prl, 'mm> {
     /// Set a specific encoder frame setting
     ///
@@ -479,7 +382,33 @@ impl<'prl, 'mm> JxlEncoder<'prl, 'mm> {
         height: u32,
     ) -> Result<MultiFrames<'enc, 'prl, 'mm, U>, EncodeError> {
         self.setup_encoder(width, height, U::bits_per_sample(), self.has_alpha)?;
-        Ok(MultiFrames::<'enc, 'prl, '_, U>(self, PhantomData))
+        Ok(MultiFrames::<'enc, 'prl, 'mm, U>(self, PhantomData))
+    }
+
+    /// Add a metadata box to the encoder
+    ///
+    /// # Errors
+    /// Return [`EncodeError`] if it fails to add metadata
+    pub fn add_metadata(&mut self, metadata: &Metadata, compress: bool) -> Result<(), EncodeError> {
+        let (&t, &data) = match metadata {
+            Metadata::Exif(data) => (b"Exif", data),
+            Metadata::Xmp(data) => (b"xml ", data),
+            Metadata::Jumb(data) => (b"jumb", data),
+            Metadata::Custom(t, data) => (t, data),
+        };
+        if !self.use_box {
+            self.check_enc_status(unsafe { JxlEncoderUseBoxes(self.enc) })?;
+            self.use_box = true;
+        }
+        self.check_enc_status(unsafe {
+            JxlEncoderAddBox(
+                self.enc,
+                &Metadata::box_type(t),
+                data.as_ptr().cast(),
+                data.len(),
+                compress.into(),
+            )
+        })
     }
 
     /// Encode a JPEG XL image from existing raw JPEG data
@@ -510,7 +439,8 @@ impl<'prl, 'mm> JxlEncoder<'prl, 'mm> {
 
     /// Encode a JPEG XL image from pixels
     ///
-    /// Note: Use RGB(3) channels, native endianness and no alignment. Ignore alpha channel settings
+    /// Note: Use RGB(3) channels, native endianness and no alignment.
+    /// Ignore alpha channel settings
     ///
     /// # Errors
     /// Return [`EncodeError`] if the internal encoder fails to encode
@@ -525,7 +455,8 @@ impl<'prl, 'mm> JxlEncoder<'prl, 'mm> {
         self.start_encoding::<U>()
     }
 
-    /// Encode a JPEG XL image from a frame. See [`EncoderFrame`] for custom options of the original pixels.
+    /// Encode a JPEG XL image from a frame.
+    /// See [`EncoderFrame`] for custom options of the original pixels.
     ///
     /// # Errors
     /// Return [`EncodeError`] if the internal encoder fails to encode
@@ -547,46 +478,17 @@ impl Drop for JxlEncoder<'_, '_> {
     }
 }
 
-/// A wrapper type for encoding multiple frames
-pub struct MultiFrames<'enc, 'prl, 'mm, U>(&'enc mut JxlEncoder<'prl, 'mm>, PhantomData<U>)
-where
-    'prl: 'enc,
-    'mm: 'enc;
-
-impl<U: PixelType> MultiFrames<'_, '_, '_, U> {
-    /// Add a frame to the encoder
-    /// # Errors
-    /// Return [`EncodeError`] if the internal encoder fails to add a frame
-    pub fn add_frame<T: PixelType>(self, frame: &EncoderFrame<T>) -> Result<Self, EncodeError> {
-        self.0.add_frame(frame)?;
-        Ok(self)
-    }
-
-    /// Add a JPEG raw frame to the encoder
-    /// # Errors
-    /// Return [`EncodeError`] if the internal encoder fails to add a jpeg frame
-    pub fn add_jpeg_frame(self, data: &[u8]) -> Result<Self, EncodeError> {
-        self.0.add_jpeg_frame(data)?;
-        Ok(self)
-    }
-
-    /// Encode a JPEG XL image from the frames
-    /// # Errors
-    /// Return [`EncodeError`] if the internal encoder fails to encode
-    pub fn encode(self) -> Result<EncoderResult<U>, EncodeError> {
-        self.0.start_encoding()
-    }
-}
-
 /// Return a [`JxlEncoderBuilder`] with default settings
 #[must_use]
 pub fn encoder_builder<'prl, 'mm>() -> JxlEncoderBuilder<'prl, 'mm> {
     JxlEncoderBuilder::default()
 }
 
+// MARK: Tests
 #[cfg(test)]
 mod tests {
     use super::*;
+    use testresult::TestResult;
 
     #[test]
     #[allow(clippy::clone_on_copy)]
@@ -599,4 +501,13 @@ mod tests {
 
         _ = encoder_builder().clone();
     }
+
+    #[test]
+    fn test_usebox() -> TestResult {
+        let mut encoder = encoder_builder().build()?;
+        let metadata = Metadata::Exif(&[0, 1, 2, 3]);
+        encoder.add_metadata(&metadata, true)?;
+        assert!(encoder.use_box);
+        Ok(())
+    }
 }
